from django import forms

from taggit.utils import parse_tags


class TagWidget(forms.TextInput):
    def render(self, name, value, attrs=None):
        if value is not None and not isinstance(value, basestring):
            value = ", ".join(o.tag.name for o in value.select_related("tag"))
        return super(TagWidget, self).render(name, value, attrs)

class TagField(forms.CharField):
    widget = TagWidget
    
    def clean(self, value):
        try:
            return parse_tags(value)
        except ValueError:
            raise forms.ValidationError("Please provide a comma-separated list of tags.")



import re

from django.conf import settings

from taggit.contrib.suggest.models import TagKeyword, TagRegex
from taggit.models import Tag


def _suggest_keywords(content):
    """ Suggest by keywords """
    suggested_keywords = set()
    keywords = TagKeyword.objects.all()

    for k in keywords:
        # Use the stem if available, otherwise use the whole keyword
        if k.stem:
            if k.stem in content:
                suggested_keywords.add(k.tag_id)
        elif k.keyword in content:
            suggested_keywords.add(k.tag_id)

    return suggested_keywords

def _suggest_regexes(content):
    """ Suggest by regular expressions """
    # Grab all regular expressions and compile them
    suggested_regexes = set()
    regex_keywords = TagRegex.objects.all()

    # Look for our regular expressions in the content
    for r in regex_keywords:
        if re.search(r.regex, content):
            suggested_regexes.add(r.tag_id)

    return suggested_regexes

def suggest_tags(content):
    """ Suggest tags based on text content """
    suggested_keywords = _suggest_keywords(content)
    suggested_regexes  = _suggest_regexes(content)
    suggested_tag_ids  = suggested_keywords | suggested_regexes

    return Tag.objects.filter(id__in=suggested_tag_ids)



from collections import defaultdict

import django
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models.related import RelatedObject
from django.db.models.fields.related import ManyToManyRel
from django.db.models.query_utils import QueryWrapper

from taggit.forms import TagField
from taggit.models import Tag, TaggedItem
from taggit.utils import require_instance_manager


class TaggableRel(ManyToManyRel):
    def __init__(self):
        self.to = TaggedItem
        self.related_name = None
        self.limit_choices_to = {}
        self.symmetrical = True
        self.multiple = True
        self.through = None


class TaggableManager(object):
    def __init__(self, verbose_name="Tags"):
        self.rel = TaggableRel()
        self.verbose_name = verbose_name
        self.editable = True
        self.unique = False
        self.creates_table = False
        self.db_column = None
        self.choices = None
        self.serialize = False
        self.creation_counter = models.Field.creation_counter
        models.Field.creation_counter += 1

    def __get__(self, instance, type):
        manager = _TaggableManager()
        manager.model = type
        if instance is None:
            manager.object_id = None
        elif instance.pk is None:
            raise ValueError("%s objects need to have a primary key value "
                "before you can access their tags." % type.__name__)
        else:
            manager.object_id = instance.pk
        return manager

    def contribute_to_class(self, cls, name):
        self.name = self.column = name
        self.model = cls
        cls._meta.add_field(self)
        setattr(cls, name, self)

    def save_form_data(self, instance, value):
        getattr(instance, self.name).set(*value)

    def get_prep_lookup(self, lookup_type, value):
        if lookup_type !=  "in":
            raise ValueError("You can't do lookups other than \"in\" on Tags")
        if all(isinstance(v, Tag) for v in value):
            qs = TaggedItem.objects.filter(tag__in=value)
        elif all(isinstance(v, basestring) for v in value):
            qs = TaggedItem.objects.filter(tag__name__in=value)
        elif all(isinstance(v, (int, long)) for v in value):
            # This one is really ackward, just don't do it.  The ORM does it
            # for deletes, but no one else gets to.
            return value
        else:
            # Fucking flip-floppers.
            raise ValueError("You can't combine Tag objects and strings. '%s' was provided." % value)
        if hasattr(models.Field, "get_prep_lookup"):
            return models.Field().get_prep_lookup(lookup_type, qs)
        return models.Field().get_db_prep_lookup(lookup_type, qs)
    
    if django.VERSION < (1, 2):
        get_db_prep_lookup = get_prep_lookup
    else:
        def get_db_prep_lookup(self, lookup_type, value, connection, prepared=False):
            if not prepared:
                return self.get_prep_lookup(lookup_type, value)
            return models.Field().get_db_prep_lookup(lookup_type, value, connection=connection, prepared=True)

    def formfield(self, form_class=TagField, **kwargs):
        defaults = {
            "label": "Tags",
            "help_text": "A comma-separated list of tags."
        }
        defaults.update(kwargs)
        return form_class(**kwargs)

    def value_from_object(self, instance):
        if instance.pk:
            return TaggedItem.objects.filter(
                object_id=instance.pk,
                content_type=ContentType.objects.get_for_model(instance)
            )
        return TaggedItem.objects.none()

    def related_query_name(self):
        return None

    def m2m_reverse_name(self):
        return "id"

    def m2m_column_name(self):
        return "object_id"

    def db_type(self, connection=None):
        return None

    def m2m_db_table(self):
        return self.rel.to._meta.db_table

    def extra_filters(self, pieces, pos, negate):
        if negate:
            return []
        prefix = "__".join(pieces[:pos+1])
        cts = map(ContentType.objects.get_for_model, _get_subclasses(self.model))
        if len(cts) == 1:
            return [("%s__content_type" % prefix, cts[0])]
        return [("%s__content_type__in" % prefix, cts)]


class _TaggableManager(models.Manager):
    def get_query_set(self):
        ct = ContentType.objects.get_for_model(self.model)
        if self.object_id is not None:
            return Tag.objects.filter(items__object_id=self.object_id,
                items__content_type=ct)
        else:
            return Tag.objects.filter(items__content_type=ct).distinct()

    @require_instance_manager
    def add(self, *tags):
        for tag in tags:
            if not isinstance(tag, Tag):
                tag, _ = Tag.objects.get_or_create(name=tag)
            TaggedItem.objects.get_or_create(object_id=self.object_id,
                content_type=ContentType.objects.get_for_model(self.model), tag=tag)

    @require_instance_manager
    def set(self, *tags):
        self.clear()
        self.add(*tags)

    @require_instance_manager
    def remove(self, *tags):
        TaggedItem.objects.filter(object_id=self.object_id,
            content_type=ContentType.objects.get_for_model(self.model)).filter(
            tag__name__in=tags).delete()

    @require_instance_manager
    def clear(self):
        TaggedItem.objects.filter(object_id=self.object_id,
            content_type=ContentType.objects.get_for_model(self.model)).delete()

    def most_common(self):
        return self.get_query_set().annotate(
            num_times=models.Count('items')
        ).order_by('-num_times')

    @require_instance_manager
    def similar_objects(self):
        qs = TaggedItem.objects.values('object_id', 'content_type')
        qs = qs.annotate(n=models.Count('pk'))
        qs = qs.exclude(
            object_id=self.object_id,
            content_type=ContentType.objects.get_for_model(self.model)
        )
        qs = qs.filter(tag__in=self.all())
        qs = qs.order_by('-n')

        preload = defaultdict(set)
        for result in qs:
            preload[result["content_type"]].add(result["object_id"])

        items = {}
        for ct, obj_ids in preload.iteritems():
            ct = ContentType.objects.get_for_id(ct)
            items[ct.pk] = dict((o.pk, o) for o in
                ct.model_class()._default_manager.filter(pk__in=obj_ids)
            )

        results = []
        for result in qs:
            obj = items[result["content_type"]][result["object_id"]]
            obj.similar_tags = result["n"]
            results.append(obj)
        return results

def _get_subclasses(model):
    subclasses = [model]
    for f in model._meta.get_all_field_names():
        field = model._meta.get_field_by_name(f)[0]
        if isinstance(field, RelatedObject) and getattr(field.field.rel, "parent_link", None):
            subclasses.extend(_get_subclasses(field.model))
    return subclasses



import re

from django.core.exceptions import ValidationError
from django.db import models

from taggit.models import Tag

try:
    import Stemmer
except ImportError:
    Stemmer = None


class TagKeyword(models.Model):
    """ Model to associate simple keywords to a Tag """
    tag = models.ForeignKey(Tag, related_name='keywords')
    keyword = models.CharField(max_length=30)
    stem = models.CharField(max_length=30)

    def __unicode__(self):
        return "Keyword '%s' for Tag '%s'" % (self.keyword, self.tag.name)

    def save(self, *args, **kwargs):
        """ Stem the keyword on save if they have PyStemmer """
        language = kwargs.pop('stemmer-language', 'english')
        if not self.pk and not self.stem and Stemmer:
            stemmer = Stemmer.Stemmer(language)
            self.stem = stemmer.stemWord(self.keyword)
        super(TagKeyword, self).save(*args, **kwargs)


def validate_regex(value):
    """ Make sure we have a valid regular expression """
    try:
        re.compile(value)
    except Exception:
        raise ValidationError('Please enter a valid regular expression')


class TagRegex(models.Model):
    """ Model to associate regular expressions with a Tag """
    tag = models.ForeignKey(Tag, related_name='regexes')
    name = models.CharField(max_length=30)
    regex = models.CharField(
        max_length=250,
         validators=[validate_regex],
         help_text=('Enter a valid Regular Expression. To make it '
            'case-insensitive include "(?i)" in your expression.')
     )

    def __unicode__(self):
        return self.name

    def save(self, *args, **kwargs):
        """ Make sure to validate """
        self.full_clean()
        super(TagRegex,self).save(*args, **kwargs)



from django.contrib import admin

from taggit.models import Tag, TaggedItem


class TaggedItemInline(admin.StackedInline):
    model = TaggedItem

class TagAdmin(admin.ModelAdmin):
    inlines = [
        TaggedItemInline
    ]


admin.site.register(Tag, TagAdmin)



from django.contrib import admin

from taggit.admin import TaggedItemInline
from taggit.contrib.suggest.models import TagKeyword, TagRegex
from taggit.models import Tag


class TagKeywordInline(admin.StackedInline):
    model = TagKeyword


class TagRegxInline(admin.StackedInline):
    model = TagRegex


class TagSuggestAdmin(admin.ModelAdmin):
    inlines = [
        TaggedItemInline,
        TagKeywordInline,
        TagRegxInline,
    ]


admin.site.unregister(Tag)
admin.site.register(Tag, TagSuggestAdmin)



from django.core.exceptions import ValidationError
from django.test import TestCase

from taggit.contrib.suggest.models import TagKeyword, TagRegex
from taggit.contrib.suggest.utils import suggest_tags
from taggit.models import Tag


class SuggestCase(TestCase):
    def test_simple_suggest(self):
        ku_tag = Tag.objects.create(name='ku')
        ku_keyword1 = TagKeyword.objects.create(
            tag=ku_tag,
            keyword='kansas university'
        )

        suggested_tags = suggest_tags('I used to be a student at kansas university')
        self.assertTrue(ku_tag in suggested_tags)

    def test_regex_suggest(self):
        ku_tag = Tag.objects.create(name='ku')
        TagRegex.objects.create(
            tag=ku_tag,
            name='Find University of Kansas',
            regex='University\s+of\s+Kansas'
        )

        suggested_tags = suggest_tags('I was once a student at the University of Kansas')

        self.assertTrue(ku_tag in suggested_tags)

    def test_bad_regex(self):
        ku_tag = Tag.objects.create(name='ku')
        ku_keyword1 = TagKeyword.objects.create(
            tag=ku_tag,
            keyword='kansas university'
        )
        new_regex = TagRegex(
            tag=ku_tag,
            name='Find University of Kansas',
            regex='University\s+of(\s+Kansas'
        )
        self.assertRaises(ValidationError, new_regex.save)

        suggested_tags = suggest_tags('I was once a student at the University '
            'of Kansas. Also known as kansas university by the way.')

        self.assertTrue(ku_tag in suggested_tags)



from functools import wraps

def parse_tags(tags):
    if tags is None:
        tags = ""
    return [o.strip() for o in tags.split(',') if o.strip()]

def require_instance_manager(func):
    @wraps(func)
    def inner(self, *args, **kwargs):
        if self.object_id is None:
            raise TypeError("Can't call %s with a non-instance manager" % func.__name__)
        return func(self, *args, **kwargs)
    return inner



from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.generic import GenericForeignKey
from django.db import models, IntegrityError
from django.template.defaultfilters import slugify


class Tag(models.Model):
    name = models.CharField(max_length=100)
    slug = models.SlugField(unique=True, max_length=100)
    
    def __unicode__(self):
        return self.name
    
    def save(self, *args, **kwargs):
        if not self.pk and not self.slug:
            self.slug = slug = slugify(self.name)
            i = 0
            while True:
                try:
                    return super(Tag, self).save(*args, **kwargs)
                except IntegrityError:
                    i += 1
                    self.slug = "%s_%d" % (slug, i)
        else:
            return super(Tag, self).save(*args, **kwargs)


class TaggedItem(models.Model):
    object_id = models.IntegerField()
    content_type = models.ForeignKey(ContentType, related_name="tagged_items")
    content_object = GenericForeignKey()
    
    tag = models.ForeignKey(Tag, related_name="items")
    
    def __unicode__(self):
        return "%s tagged with %s" % (self.content_object, self.tag)



from __future__ import with_statement
from contextlib import contextmanager

from django.test import TestCase

from taggit.models import Tag
from taggit.tests.forms import FoodForm
from taggit.tests.models import Food, Pet, HousePet


class BaseTaggingTest(TestCase):
    def assert_tags_equal(self, qs, tags, sort=True):
        got = map(lambda tag: tag.name, qs)
        if sort:
            got.sort()
            tags.sort()
        self.assertEqual(got, tags)
    
    @contextmanager
    def assert_raises(self, exc_type):
        try:
            yield
        except Exception, e:
            self.assert_(type(e) is exc_type, "%s didn't match expected "
                "exception type %s" % (e, exc_type))
        else:
            self.fail("No exception raised, expected %s" % exc_type)
    

class AddTagTestCase(BaseTaggingTest):
    def test_add_tag(self):
        apple = Food.objects.create(name="apple")
        self.assertEqual(list(apple.tags.all()), [])
        self.assertEqual(list(Food.tags.all()),  [])

        apple.tags.add('green')
        self.assert_tags_equal(apple.tags.all(), ['green'])
        self.assert_tags_equal(Food.tags.all(), ['green'])

        pear = Food.objects.create(name="pear")
        pear.tags.add('green')
        self.assert_tags_equal(pear.tags.all(), ['green'])
        self.assert_tags_equal(Food.tags.all(), ['green'])

        apple.tags.add('red')
        self.assert_tags_equal(apple.tags.all(), ['green', 'red'])
        self.assert_tags_equal(Food.tags.all(), ['green', 'red'])

        self.assert_tags_equal(Food.tags.most_common(), ['green', 'red'], sort=False)

        apple.tags.remove('green')
        self.assert_tags_equal(apple.tags.all(), ['red'])
        self.assert_tags_equal(Food.tags.all(), ['green', 'red'])
        tag = Tag.objects.create(name="delicious")
        apple.tags.add(tag)
        self.assert_tags_equal(apple.tags.all(), ["red", "delicious"])
        
        apple.delete()
        self.assert_tags_equal(Food.tags.all(), ["green"])
        
        f = Food()
        with self.assert_raises(ValueError):
            f.tags.all()
    
    def test_unique_slug(self):
        apple = Food.objects.create(name="apple")
        apple.tags.add("Red", "red")


class DeleteObjecTestCase(BaseTaggingTest):
    def test_delete_obj(self):
        apple = Food.objects.create(name="apple")
        apple.tags.add("red")
        self.assert_tags_equal(apple.tags.all(), ["red"])
        strawberry = Food.objects.create(name="strawberry")
        strawberry.tags.add("red")
        apple.delete()
        self.assert_tags_equal(strawberry.tags.all(), ["red"])


class LookupByTagTestCase(BaseTaggingTest):
    def test_lookup_by_tag(self):
        apple = Food.objects.create(name="apple")
        apple.tags.add("red", "green")
        pear = Food.objects.create(name="pear")
        pear.tags.add("green")
        
        self.assertEqual(list(Food.objects.filter(tags__in=["red"])), [apple])
        self.assertEqual(list(Food.objects.filter(tags__in=["green"])), [apple, pear])

        kitty = Pet.objects.create(name="kitty")
        kitty.tags.add("fuzzy", "red")
        dog = Pet.objects.create(name="dog")
        dog.tags.add("woof", "red")
        self.assertEqual(list(Food.objects.filter(tags__in=["red"]).distinct()), [apple])

        tag = Tag.objects.get(name="woof")
        self.assertEqual(list(Pet.objects.filter(tags__in=[tag])), [dog])

        cat = HousePet.objects.create(name="cat", trained=True)
        cat.tags.add("fuzzy")

        self.assertEqual(
            map(lambda o: o.pk, Pet.objects.filter(tags__in=["fuzzy"])),
            [kitty.pk, cat.pk]
        )


class TaggableFormTestCase(BaseTaggingTest):
    def test_form(self):
        self.assertEqual(FoodForm.base_fields.keys(), ['name', 'tags'])

        f = FoodForm({'name': 'apple', 'tags': 'green, red, yummy'})
        self.assertEqual(str(f), """<tr><th><label for="id_name">Name:</label></th><td><input id="id_name" type="text" name="name" value="apple" maxlength="50" /></td></tr>\n<tr><th><label for="id_tags">Tags:</label></th><td><input type="text" name="tags" value="green, red, yummy" id="id_tags" /></td></tr>""")
        f.save()
        apple = Food.objects.get(name='apple')
        self.assert_tags_equal(apple.tags.all(), ['green', 'red', 'yummy'])

        f = FoodForm({'name': 'apple', 'tags': 'green, red, yummy, delicious'}, instance=apple)
        f.save()
        apple = Food.objects.get(name='apple')
        self.assert_tags_equal(apple.tags.all(), ['green', 'red', 'yummy', 'delicious'])
        self.assertEqual(Food.objects.count(), 1)
        
        f = FoodForm({"name": "raspberry"})
        raspberry = f.save()
        self.assert_tags_equal(raspberry.tags.all(), [])
        
        f = FoodForm(instance=apple)
        self.assertEqual(str(f), """<tr><th><label for="id_name">Name:</label></th><td><input id="id_name" type="text" name="name" value="apple" maxlength="50" /></td></tr>\n<tr><th><label for="id_tags">Tags:</label></th><td><input type="text" name="tags" value="green, red, yummy, delicious" id="id_tags" /></td></tr>""")


class SimilarityByTagTestCase(BaseTaggingTest):
    def test_similarity_by_tag(self):
        """Test that pears are more similar to apples than watermelons"""
        apple = Food.objects.create(name="apple")
        apple.tags.add("green", "juicy", "small", "sour")

        pear = Food.objects.create(name="pear")
        pear.tags.add("green", "juicy", "small", "sweet")

        watermelon = Food.objects.create(name="watermelon")
        watermelon.tags.add("green", "juicy", "large", "sweet")

        similar_objs = apple.tags.similar_objects()
        self.assertEqual(similar_objs, [pear, watermelon])
        self.assertEqual(map(lambda x: x.similar_tags, similar_objs), [3, 2])


class TagReuseTestCase(BaseTaggingTest):
    def test_tag_reuse(self):
        apple = Food.objects.create(name="apple")
        apple.tags.add("juicy", "juicy")
        self.assert_tags_equal(apple.tags.all(), ['juicy'])



import os
from setuptools import setup, find_packages

from taggit import VERSION


f = open(os.path.join(os.path.dirname(__file__), 'README.txt'))
readme = f.read()
f.close()

setup(
    name='django-taggit',
    version=".".join(map(str, VERSION)),
    description='django-taggit is a reusable Django application for simple tagging.',
    long_description=readme,
    author='Alex Gaynor',
    author_email='alex.gaynor@gmail.com',
    url='http://github.com/alex/django-taggit/tree/master',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Framework :: Django',
    ],
)




from django.db import models

from taggit.managers import TaggableManager


class Food(models.Model):
    name = models.CharField(max_length=50)
    
    tags = TaggableManager()
    
    def __unicode__(self):
        return self.name


class Pet(models.Model):
    name = models.CharField(max_length=50)
    
    tags = TaggableManager()

class HousePet(Pet):
    trained = models.BooleanField()